-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all; 
use work.sha3_pkg.all;
use work.keccak_pkg.all;

-- Possible generics values: 
-- hs = {HASH_SIZE_256, HASH_SIZE_512} 

entity keccak_control is
	port (					
		clk					: in std_logic;
		rst					: in std_logic;	
		
		-- fifo
		src_ready			: in std_logic;
		src_read			: out std_logic;
		dst_ready 			: in std_logic;
		dst_write			: out std_logic;
								
		-- input
		loadLen				:out std_logic;
		decLen				:out std_logic;	
		c					:in std_logic_vector(31 downto 0); 
		final_segment 		:in std_logic;
		-- process						
		selCI				:out std_logic;
		selTRP				:out std_logic;		
		selLastByte 		:out std_logic;	 
		enSaveRotatedRho	:out std_logic;

		selWriteShift 		:out std_logic;
		selAbsorb			:out std_logic;
		rstMem				:out std_logic;	  
		selWriteMem			:out std_logic;
		selReadMem			:out std_logic;
		writeLane			:out std_logic_vector(4 downto 0);	
		rc_addr				:out std_logic_vector(7 downto 0)		
	);
end keccak_control;

architecture h256 of keccak_control is	
	constant roundnr : integer := 24;
	
	type state_type is ( 
		reset, resetMemory, waitForHeader1State, waitForHeader2State, 
		loadLane0State, loadLane1State, loadLane2State, loadLane3State, 
		calcSumState, squeezeTRPState, squeezeState, outputState );
	signal cs, nstate : state_type;
	
	signal readByte : std_logic_vector(2 downto 0);
	signal roundCount : std_logic_vector(4 downto 0);
	signal enRoundCnt : std_logic;				   
	signal enRdByteCnt : std_logic;		 
	
	signal compReadByteEnd, compRoundCountEnd, compReadByteBegin : std_logic; 
	signal rstCtr, rstSelRdBank, rstSelWrBank, rstLastSegment : std_logic;								
	
	signal selMemWrBank, toggleMemWrBank : std_logic;
	signal selMemRdBank, toggleMemRdBank : std_logic;
	
	signal lastSegment, setLastSegment : std_logic;
begin						   
	rc_addr <= roundCount & readByte; 
	
	-- Round Counter
	roundCounterUnit: process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' or rstCtr = '1'  then
				roundCount <= (others => '0' );
			elsif enRoundCnt = '1' then
				roundCount <= roundCount + 1;
			end if;
		end if;
	end process;
			
	-- Address Counter ( Byte Selection )
	readByteCounterUnit: process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' or rstCtr = '1' then
				readByte <= (others => '0' );
			elsif enRdByteCnt = '1' then
				readByte <= readByte + 1;			
			end if;
		end if;
	end process;		
	
	-- Last Segment Register 
	lastSegmentReg: process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' or rstLastSegment = '1' then
				lastSegment <= '0';
			elsif setLastSegment = '1' then
				lastSegment <= '1';
			end if;
		end if;
	end process;		  
	
	-- Select Write Mem Toggler
	selWriteMemToggler: process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' or rstSelWrBank = '1' then
				selMemWrBank <= '0';
			elsif toggleMemWrBank = '1' then
				selMemWrBank <= not selMemWrBank;
			end if;
		end if;
	end process;	
	selWriteMem <= selMemWrBank;
	
	-- Select Read Mem Toggler
	selReadMemToggler: process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' or rstSelRdBank = '1' then
				selMemRdBank <= '0';
			elsif toggleMemRdBank = '1' then
				selMemRdBank <= not selMemRdBank;
			end if;
		end if;
	end process;	
	selReadMem <= selMemRdBank;
					
										
	-- Comparators
	compReadByteEnd <= '1' when readByte = "111" else '0';
	compRoundCountEnd <= '1' when roundCount = roundnr-1 else '0';
	compReadByteBegin <= '1' when readByte = "000" else '0';
	
	
	
	-- CSTATE Register
	cstateReg : process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' then
				cs <= reset;
			else
				cs <= nstate;
			end if;
		end if;
	end process;
			  
	-- NSTATE FSM			  
	nstateFSM : process( cs, src_ready, final_segment, lastSegment, dst_ready, c, compReadByteEnd, compReadByteBegin, compRoundCountEnd, roundCount, lastSegment ) 
	begin						  
		case cs is
			when reset =>
				nstate <= resetMemory;
			when resetMemory => 
				if ( compReadByteEnd = '1' ) then
					nstate <= waitForHeader1State;	
				else
					nstate <= resetMemory;
				end if;				
			when waitForHeader1State =>   
				if src_ready = '0' then
					if final_segment = '1' then
						nstate <= waitForHeader2State;	
					else
						nstate <= loadLane0State;
					end if;				
				else
					nstate <= waitForHeader1State;
				end if;
			when waitForHeader2State => 				  
				if src_ready = '0' then
					nstate <= loadLane0State;	
				else
					nstate <= waitForHeader2State;
				end if;				
			when loadLane0State =>
				if compReadByteEnd = '1' then
					nstate <= loadLane1State;
				else
					nstate <= loadLane0State;
				end if;
			when loadLane1State =>
				if compReadByteEnd = '1' then
					nstate <= loadLane2State; 
				else
					nstate <= loadLane1State;
				end if;
			when loadLane2State =>
				if compReadByteEnd = '1' then
					nstate <= loadLane3State;
				else
					nstate <= loadLane2State;
				end if;
			when loadLane3State =>
				if compReadByteEnd = '1' then
					nstate <= calcSumState;		
				else
					nstate <= loadLane3State;
				end if;				
			when calcSumState =>
				nstate <= squeezeTRPState;	
			when squeezeTRPState =>
				if compReadByteEnd = '1' then
					nstate <= squeezeState;
				else
					nstate <= squeezeTRPState;
				end if;
			when squeezeState => 
				if compReadByteEnd = '1' and compRoundCountEnd = '1' then
					if lastSegment = '1' and c = 0 then
						nstate <= outputState;	   
					elsif lastSegment = '0' and c = 0 then
						nstate <= waitForHeader1State;
					else
						nstate <= loadLane0State;
					end if;
				else
					nstate <= squeezeState;
				end if;
			when outputState => 	   
				if roundCount = 3 and compReadByteEnd = '1' then
					nstate <= reset;
				else
					nstate <= outputState; 
				end if;
		end case;
	end process;	
	 	
	-- Outputs Transition
	outputFSM : process( cs, src_ready, final_segment, dst_ready, c, compReadByteEnd, compReadByteBegin, roundCount, compRoundCountEnd, readByte, lastSegment )
	begin			
		-- external
		src_read <= '0'; dst_write <= '0';
		
		-- datapath
		loadLen <= '0'; decLen <= '0'; toggleMemWrBank <= '0'; toggleMemRdBank <= '0';	  
		selCI <= '0'; selTRP <= '0'; enSaveRotatedRho <= '0'; selLastByte <= '0';-- Round controls
		selAbsorb <= '0'; rstMem <= '0';   writeLane <= (others => '0'); selWriteShift <= '0';
		
		-- internal
		rstCtr <= '0'; rstSelWrBank <= '0'; rstSelRdBank <= '0'; enRoundCnt <= '0'; enRdByteCnt <= '0'; setLastSegment <= '0';
		rstLastSegment <= '0';
		
		case cs is
			when reset =>				
				rstSelWrBank <= '1';
				rstSelRdBank <= '1';
				rstCtr <= '1';		   
				rstLastSegment <= '1';
			when resetMemory =>
				rstMem <= '1';
				writeLane <= (others => '1');
				enRdByteCnt <= '1';
				selAbsorb <= '1';
			when waitForHeader1State =>
				if src_ready = '0' then
					loadLen <= '1';
					src_read <= '1';  
					if final_segment = '1' then
						setLastSegment <= '1';
					end if;
				end if;
			when waitForHeader2State => 
				if src_ready = '0' then
					src_read <= '1';					
				end if;
			when loadLane0State =>	   		   	  
				selAbsorb <= '1';
				if src_ready = '0' then
					src_read <= '1';
					enRdByteCnt <= '1';	 
					writeLane(0) <= '1';
				end if;
			when loadLane1State =>	
				selAbsorb <= '1';
				if src_ready = '0' then
					src_read <= '1';
					enRdByteCnt <= '1';	 
					writeLane(1) <= '1';
				end if;
			when loadLane2State =>	 
				selAbsorb <= '1';
				if src_ready = '0' then
					src_read <= '1';
					enRdByteCnt <= '1';	 
					writeLane(2) <= '1';
				end if;
			when loadLane3State =>	  
				selAbsorb <= '1';
				if src_ready = '0' then
					src_read <= '1';					
					writeLane(3) <= '1'; 
					if compReadByteEnd = '0' then	-- Keep the next address to be 7 so we can read the last address to precalculate the rotated sum value
						enRdByteCnt <= '1';	 	 								
					end if;					
				end if;													
			when calcSumState =>	
			   	enRdByteCnt <= '1';	 -- Make the next address to be 0
				decLen <= '1'; 		
				toggleMemWrBank <= '1';	 
			when squeezeTRPState =>
				enRdByteCnt <= '1';
				writeLane <= (others => '1');
				selTRP <= '1'; 
				selCI <= '0';
				selWriteShift <= '1';
				
				if ( compReadByteEnd = '1' ) then 				
					toggleMemWrBank <= '1';
					toggleMemRdBank <= '1';	  
					selLastByte <= '1';
				end if;				
			when squeezeState =>   		 	
				enRdByteCnt <= '1';	 

				---- Don't write at the first clock cycle of the squeeze as the data is not yet correct
				writeLane <= (others => '1');
				
				if ( compRoundCountEnd = '1' ) then	-- Only CI
					selTRP <= '0';
					selCI <= '1';
				else  -- Both
					selTRP <= '1';
					selCI <= '1';
					selWriteShift <= '1';
				end if;
				
				if ( compReadByteBegin = '1' ) then	
					enSaveRotatedRho <= '1';					
				end if;										
				
				if ( compReadByteEnd = '1' ) then 
					enRoundCnt <= '1';		 
					selLastByte <= '1';
					toggleMemRdBank <= '1';	
					if ( compRoundCountEnd = '0' ) then
						toggleMemWrBank <= '1';						
					end if;
				end if;			 			  
								
				
				if compReadByteEnd = '1' and compRoundCountEnd = '1' then
					rstCtr <= '1';
				end if;
					
			when outputState =>			 
				rstLastSegment <= '1';
				if dst_ready = '0' then	 
					dst_write <= '1';
					if ( compReadByteEnd = '1' ) then
						enRoundCnt <= '1';
					end if;
					enRdByteCnt <= '1';   
				end if;
		end case;
				
	end process;
end h256;
 
 
 --==================================================
 --==================== H512 ========================
 --==================================================
 
			 		  
architecture h512 of keccak_control is	
	constant roundnr : integer := 24;
	
	type state_type is ( 
		reset, resetMemory, waitForHeader1State, waitForHeader2State, 
		loadLane0State, loadLane1State, 
		calcSumState, squeezeTRPState, squeezeState, outputState );
	signal cs, nstate : state_type;
	
	signal readByte : std_logic_vector(2 downto 0);
	signal roundCount : std_logic_vector(4 downto 0);
	signal enRoundCnt : std_logic;				   
	signal enRdByteCnt : std_logic;		 
	
	signal compReadByteEnd, compRoundCountEnd, compReadByteBegin : std_logic; 
	signal rstCtr, rstSelRdBank, rstSelWrBank, rstLastSegment : std_logic;								
	
	signal selMemWrBank, toggleMemWrBank : std_logic;
	signal selMemRdBank, toggleMemRdBank : std_logic;
	
	signal lastSegment, setLastSegment : std_logic;
begin						   
	rc_addr <= roundCount & readByte; 
	
	-- Round Counter
	roundCounterUnit: process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' or rstCtr = '1'  then
				roundCount <= (others => '0' );
			elsif enRoundCnt = '1' then
				roundCount <= roundCount + 1;
			end if;
		end if;
	end process;
			
	-- Address Counter ( Byte Selection )
	readByteCounterUnit: process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' or rstCtr = '1' then
				readByte <= (others => '0' );
			elsif enRdByteCnt = '1' then
				readByte <= readByte + 1;			
			end if;
		end if;
	end process;		
	
	-- Last Segment Register 
	lastSegmentReg: process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' or rstLastSegment = '1' then
				lastSegment <= '0';
			elsif setLastSegment = '1' then
				lastSegment <= '1';
			end if;
		end if;
	end process;		  
	
	-- Select Write Mem Toggler
	selWriteMemToggler: process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' or rstSelWrBank = '1' then
				selMemWrBank <= '0';
			elsif toggleMemWrBank = '1' then
				selMemWrBank <= not selMemWrBank;
			end if;
		end if;
	end process;	
	selWriteMem <= selMemWrBank;
	
	-- Select Read Mem Toggler
	selReadMemToggler: process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' or rstSelRdBank = '1' then
				selMemRdBank <= '0';
			elsif toggleMemRdBank = '1' then
				selMemRdBank <= not selMemRdBank;
			end if;
		end if;
	end process;	
	selReadMem <= selMemRdBank;
					
										
	-- Comparators
	compReadByteEnd <= '1' when readByte = "111" else '0';
	compRoundCountEnd <= '1' when roundCount = roundnr-1 else '0';
	compReadByteBegin <= '1' when readByte = "000" else '0';
	
	
	
	-- CSTATE Register
	cstateReg : process( clk )
	begin
		if rising_edge( clk ) then
			if rst = '1' then
				cs <= reset;
			else
				cs <= nstate;
			end if;
		end if;
	end process;
			  
	-- NSTATE FSM			  
	nstateFSM : process( cs, src_ready, final_segment, lastSegment, dst_ready, c, compReadByteEnd, compReadByteBegin, compRoundCountEnd, roundCount, lastSegment ) 
	begin						  
		case cs is
			when reset =>
				nstate <= resetMemory;
			when resetMemory => 
				if ( compReadByteEnd = '1' ) then
					nstate <= waitForHeader1State;	
				else
					nstate <= resetMemory;
				end if;				
			when waitForHeader1State =>   
				if src_ready = '0' then
					if final_segment = '1' then
						nstate <= waitForHeader2State;	
					else
						nstate <= loadLane0State;
					end if;				
				else
					nstate <= waitForHeader1State;
				end if;
			when waitForHeader2State => 				  
				if src_ready = '0' then
					nstate <= loadLane0State;	
				else
					nstate <= waitForHeader2State;
				end if;				
			when loadLane0State =>
				if compReadByteEnd = '1' then
					nstate <= loadLane1State;
				else
					nstate <= loadLane0State;
				end if;
			when loadLane1State =>
				if compReadByteEnd = '1' then
					nstate <= calcSumState; 
				else
					nstate <= loadLane1State;
				end if;			
			when calcSumState =>
				nstate <= squeezeTRPState;	
			when squeezeTRPState =>
				if compReadByteEnd = '1' then
					nstate <= squeezeState;
				else
					nstate <= squeezeTRPState;
				end if;
			when squeezeState => 
				if compReadByteEnd = '1' and compRoundCountEnd = '1' then
					if lastSegment = '1' and c = 0 then
						nstate <= outputState;	   
					elsif lastSegment = '0' and c = 0 then
						nstate <= waitForHeader1State;
					else
						nstate <= loadLane0State;
					end if;
				else
					nstate <= squeezeState;
				end if;
			when outputState => 	   
				if roundCount = 7 and compReadByteEnd = '1' then
					nstate <= reset;
				else
					nstate <= outputState; 
				end if;
		end case;
	end process;	
	 	
	-- Outputs Transition
	outputFSM : process( cs, src_ready, final_segment, dst_ready, c, compReadByteEnd, compReadByteBegin, roundCount, compRoundCountEnd, readByte, lastSegment )
	begin			
		-- external
		src_read <= '0'; dst_write <= '0';
		
		-- datapath
		loadLen <= '0'; decLen <= '0'; toggleMemWrBank <= '0'; toggleMemRdBank <= '0';	  
		selCI <= '0'; selTRP <= '0'; enSaveRotatedRho <= '0'; selLastByte <= '0';-- Round controls
		selAbsorb <= '0'; rstMem <= '0';   writeLane <= (others => '0'); selWriteShift <= '0';
		
		-- internal
		rstCtr <= '0'; rstSelWrBank <= '0'; rstSelRdBank <= '0'; enRoundCnt <= '0'; enRdByteCnt <= '0'; setLastSegment <= '0';
		rstLastSegment <= '0';
		
		case cs is
			when reset =>				
				rstSelWrBank <= '1';
				rstSelRdBank <= '1';
				rstCtr <= '1';		   
				rstLastSegment <= '1';
			when resetMemory =>
				rstMem <= '1';
				writeLane <= (others => '1');
				enRdByteCnt <= '1';
				selAbsorb <= '1';
			when waitForHeader1State =>
				if src_ready = '0' then
					loadLen <= '1';
					src_read <= '1';  
					if final_segment = '1' then
						setLastSegment <= '1';
					end if;
				end if;
			when waitForHeader2State => 
				if src_ready = '0' then
					src_read <= '1';					
				end if;
			when loadLane0State =>	   		   	  
				selAbsorb <= '1';
				if src_ready = '0' then
					src_read <= '1';
					enRdByteCnt <= '1';	 
					writeLane(0) <= '1';
				end if;
			when loadLane1State =>	
				selAbsorb <= '1';
				if src_ready = '0' then
					src_read <= '1';
					writeLane(1) <= '1';
					if compReadByteEnd = '0' then	-- Keep the next address to be 7 so we can read the last address to precalculate the rotated sum value
						enRdByteCnt <= '1';	 	 								
					end if;
				end if;							
			when calcSumState =>	
			   	enRdByteCnt <= '1';	 -- Make the next address to be 0
				decLen <= '1'; 		
				toggleMemWrBank <= '1';	 
			when squeezeTRPState =>
				enRdByteCnt <= '1';
				writeLane <= (others => '1');
				selTRP <= '1'; 
				selCI <= '0';
				selWriteShift <= '1';
				
				if ( compReadByteEnd = '1' ) then 				
					toggleMemWrBank <= '1';
					toggleMemRdBank <= '1';	  
					selLastByte <= '1';
				end if;				
			when squeezeState =>   		 	
				enRdByteCnt <= '1';	 

				---- Don't write at the first clock cycle of the squeeze as the data is not yet correct
				writeLane <= (others => '1');
				
				if ( compRoundCountEnd = '1' ) then	-- Only CI
					selTRP <= '0';
					selCI <= '1';
				else  -- Both
					selTRP <= '1';
					selCI <= '1';
					selWriteShift <= '1';
				end if;
				
				if ( compReadByteBegin = '1' ) then	
					enSaveRotatedRho <= '1';					
				end if;										
				
				if ( compReadByteEnd = '1' ) then 
					enRoundCnt <= '1';		 
					selLastByte <= '1';
					toggleMemRdBank <= '1';	
					if ( compRoundCountEnd = '0' ) then
						toggleMemWrBank <= '1';						
					end if;
				end if;			 			  
								
				
				if compReadByteEnd = '1' and compRoundCountEnd = '1' then
					rstCtr <= '1';
				end if;
					
			when outputState =>			 
				rstLastSegment <= '1';
				if dst_ready = '0' then	 
					dst_write <= '1';
					if ( compReadByteEnd = '1' ) then
						enRoundCnt <= '1';
					end if;
					enRdByteCnt <= '1';   
				end if;
		end case;
				
	end process;
end h512;
 