-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.sha3_pkg.all;
use work.keccak_pkg.all;

-- Possible generics values: 
-- hs = {HASH_SIZE_256, HASH_SIZE_512} 
-- b = {KECCAK256_CAPACITY, KECCAK512_CAPACITY}
-- possible combinaations of (hs, b) = {(HASH_SIZE_256, KECCAK256_CAPACITY), (HASH_SIZE_512, KECCAK512_CAPACITY)}

entity keccak_datapath is
generic (HS: integer := HASH_SIZE_256);	
port (
		clk 				:in std_logic;
		din 				:in std_logic_vector(iw-1 downto 0);
		dout 				:out std_logic_vector(ow-1 downto 0);
		
		-- input
		loadLen				:in std_logic;
		decLen				:in std_logic;	
		c					:out std_logic_vector(31 downto 0); 
		-- process						
		selCI				:in std_logic;
		selTRP				:in std_logic;
		selLastByte 		:in std_logic;	 
		enSaveRotatedRho	:in std_logic;	  		
											
		-- selRdAddr			:in std_logic;	
		selWriteShift		:in std_logic;
		selAbsorb			:in std_logic;
		rstMem				:in std_logic;	  
		selWriteMem			:in std_logic; -- select which bank to write 
		selReadMem			:in std_logic; -- select which bank to read
		writeLane			:in std_logic_vector(4 downto 0);	
		rc_addr				:in std_logic_vector(7 downto 0)		
	);
end keccak_datapath;			   

architecture struct of keccak_datapath is  						   
	signal roundIn, roundOut : std_logic_vector(199 downto 0);						   

	signal rc	:	std_logic_vector(7  downto 0);	

	signal c_wire, c_wire_in : std_logic_vector(31 downto 0);
	 
	-- Keccak data types 						 
	type double_lane_type is array (0 to 15) of std_logic_vector(7 downto 0);
	type double_plane_type is array (0 to 4) of double_lane_type;	
	
	signal state0, state1, state2, state3, state4 : double_plane_type;	
	signal stateInPre, stateIn, stateOut : field_type;
				 
	signal rdAddr : std_logic_vectoR(3 downto 0); 	

	
	type sub_addr_type is array ( 0 to 4 ) of std_logic_vector(3 downto 0);									   
	type addr_type is array ( 0 to 4 ) of sub_addr_type;
	signal wrAddr : addr_type;
	--signal rdAddr : addr_type;
	
	type din_array_type is array (0 to 4) of std_logic_vector(7 downto 0);
	signal dinArray : din_array_type;							
	
	signal roundOutArray: field_type;
	
	type rc63rd_bit_rom_type is array (0 to 31) of std_logic;
	constant rc63rd_bit : rc63rd_bit_rom_type := (
	     '0', '1', '1',
	'0', '0', '1', '1',
	'0', '0', '0', '0',
	'0', '1', '1', '1',
	'1', '1', '0', '1',
	'1', '1', '0', '1',
	'0', '0', '0', '0',
	'0', '0', '0', '0',
	'0' );	 
	signal rcCarry : std_logic;
	
	
				   
begin				
	rdAddr <= selReadMem & rc_addr(2 downto 0);
	rcCarry <= rc63rd_bit(conv_integer(rc_addr(7 downto 3)));
	
	-- segment counter
	segment_cntr_gen : process( clk )
	begin
		if rising_edge( clk ) then
			if ( loadLen = '1' ) then
				c_wire <= din(31 downto 0);
			elsif ( decLen = '1' ) then
				c_wire <= c_wire_in;
			end if;
		end if;
	end process;				   
	GenSubH256: if HS = 256 generate
		c_wire_in <= c_wire - 1280;
	end generate;
	GenSubH512: if HS = 512 generate
		c_wire_in <= c_wire - 640;
	end generate;
	c <= c_wire;
				   
	-- =====================================================
	GenDinArray: for col in 0 to 4 generate
		dinArray(col) <= din(iw-col*8-1 downto iw-col*8-8);
	end generate;
	
	GenStateInRow: for row in 0 to 3 generate -- the last lane S(4)(x) will never absorb a message, so doesn't need additional mux
		GenStateInCol: for col in 0 to 4 generate
			stateInPre(row)(col) <= stateOut(row)(col) xor dinArray(col) when selAbsorb = '1' else roundOutArray(row)(col);
			stateIn(row)(col) <= (others => '0') when rstMem = '1' else stateInPre(row)(col);		  -- for reset
		end generate;
	end generate;																											 
	stateInPre(4) <= roundOutArray(4);		 
	genStateIn4Col: for col in 0 to 4 generate
		stateIn(4)(col) <= (others => '0') when rstMem ='1' else stateInPre(4)(col);		 
	end generate;
																			
					
	-- Write Address Calculation					
	wrAddr(0)(0) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0));
	wrAddr(0)(1) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 5);
	wrAddr(0)(2) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 5);
	wrAddr(0)(3) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 2);
	wrAddr(0)(4) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 1);
		
	wrAddr(1)(0) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 3);
	wrAddr(1)(1) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 2);
	wrAddr(1)(2) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0));
	wrAddr(1)(3) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 5);
	wrAddr(1)(4) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 7);
	
	wrAddr(2)(0) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0));
	wrAddr(2)(1) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0));
	wrAddr(2)(2) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 3);
	wrAddr(2)(3) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 1);
	wrAddr(2)(4) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 2);
																			 
	wrAddr(3)(0) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 3);
	wrAddr(3)(1) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 4);
	wrAddr(3)(2) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 1);
	wrAddr(3)(3) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 1);
	wrAddr(3)(4) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 7);
																			 
	wrAddr(4)(0) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 7);
	wrAddr(4)(1) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 6);
	wrAddr(4)(2) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 4);
	wrAddr(4)(3) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0) + 5);
	wrAddr(4)(4) <= selWriteMem & rc_addr(2 downto 0) when selWriteShift = '0' else selWriteMem & (rc_addr(2 downto 0));	  
	
	-- memory	(5 row x 5 col x 16 location x 8 bits)	
	-- Note: Cannot perform an operation on a 3d array in Virtex 6 for xilinx v13.1. Resorted to declare 5 2d arrays instead.
	MemoryUnit: process( clk ) 
	begin
		if rising_edge( clk ) then			
			for col in 0 to 4 loop  				
				if (writeLane(0) = '1') then
					state0(col)(conv_integer(wrAddr(0)(col))) <= stateIn(0)(col);
				end if;	
				if (writeLane(1) = '1') then
					state1(col)(conv_integer(wrAddr(1)(col))) <= stateIn(1)(col);
				end if;	
				if (writeLane(2) = '1') then
					state2(col)(conv_integer(wrAddr(2)(col))) <= stateIn(2)(col);
				end if;	
				if (writeLane(3) = '1') then					
					state3(col)(conv_integer(wrAddr(3)(col))) <= stateIn(3)(col);
				end if;	
				if (writeLane(4) = '1') then					
					state4(col)(conv_integer(wrAddr(4)(col))) <= stateIn(4)(col);										
				end if;	
			end loop;			
		end if;
	end process;		
	
	GenStateOutCol: for col in 0 to 4 generate
		stateOut(0)(col) <= state0(col)(conv_integer(rdAddr));
		stateOut(1)(col) <= state1(col)(conv_integer(rdAddr));
		stateOut(2)(col) <= state2(col)(conv_integer(rdAddr));
		stateOut(3)(col) <= state3(col)(conv_integer(rdAddr));
		stateOut(4)(col) <= state4(col)(conv_integer(rdAddr));
	end generate;
								   
	-- asynchronous memory for Keccak constants	
	rd_cons 	: entity work.keccak_cons(keccak_cons) 	port map (addr=>rc_addr, rc=>rc);
	-- Keccak round function with architecture based on Marcin Rogawski implementation
	rd 			: entity work.keccak_fv8_round(structure)
		port map ( clk => clk, rc=>rc, 
					selCI => selCI, selTRP => selTRP, selLastByte => selLastByte, enSaveRotatedRho => enSaveRotatedRho,	byteAddr => rc_addr(2 downto 0),			
					rcCarry => rcCarry, rin=>roundIn, rout=>roundOut);
	
	roundIn <= field2str(stateOut);
	roundOutArray <= str2field(roundOut);  
	
	H256Gen: if ( HS = 256 ) generate
		with rc_addr(4 downto 3) select
		dout <= stateOut(0)(0) when "00",
				stateOut(0)(1) when "01",
				stateOut(0)(2) when "10",
				stateOut(0)(3) when others;
	end generate;
	
	H512Gen: if ( HS = 512 ) generate
		with rc_addr(5 downto 3) select
		dout <= stateOut(0)(0) when "000",
				stateOut(0)(1) when "001",
				stateOut(0)(2) when "010",
				stateOut(0)(3) when "011",
				stateOut(0)(4) when "100",
				stateOut(1)(0) when "101",				
				stateOut(1)(1) when "110",
				stateOut(1)(2) when others;
	end generate;
end struct;